#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2022-2023  Nexedi SA and Contributors.
#                          Started by Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Program nxd-bom generates bill of material for a software-release or slapos node.

Usage: nxd-bom software   <path-to-installed-software>
       nxd-bom node       <slapos deploy-script>

An example of generated bill of material is provided in example/ors-bom.txt .
"""

from __future__ import print_function

import sys, configparser, re
from os.path import basename
from glob import glob
from collections import namedtuple
from urllib.parse import unquote


# PkgInfo represents information about a package
PkgInfo = namedtuple('PkgInfo', [
    'name',
    'version',
    'kind',
    'url'])
    # TODO patches



# bom_software retrieves BOM from .installed.cfg generated by buildout along the build.
def bom_software(installed_software_path): # -> {} name -> PkgInfo
    bom = {}
    def addbom(urlpath, kind):
        name, ver = namever(urlpath)
        if '//' in urlpath:
            url = urlpath
        else:
            if kind == 'egg':
                # XXX not strictly correct -> better retrieve the actual URL, but buildout does not save it in installed.cfg
                url = 'https://pypi.org/project/%s/%s/' % (name, ver)
            else:
                raise NotImplementedError('TODO url for kind %s' % kind)


        info = PkgInfo(name, ver, kind, url)
        if name in bom:
            assert bom[name] == info,  (bom[name], info)
        else:
            bom[name] = info


    idb = configparser.ConfigParser()
    idb.read('%s/.installed.cfg' % installed_software_path)


    for s in idb.sections():
        if s == 'buildout':
            continue    # [buildout] is used internally

        part = idb[s]
        recipe = part['recipe']

        if recipe == 'slapos.recipe.cmmi':
            addbom(geturl(part), '')    # XXX detect kind?

        elif recipe == 'slapos.recipe.build':
            # slapos.recipe.build is often used in creative ways to actually
            # run python code during the build. Let's detect this via lack of
            # `url`, but be careful and explicitly list name of such sections
            # not to miss a normal part.
            url = geturl(part, None)
            if url is None:
                # [gcc] is generic section to either use system gcc, or install one via slapos
                # the one installed via slapos will come in its own section
                # [python] is similar - it picks up  python2 or python3
                # ZODB, ZEO, ... are similar as well
                if s in ('gcc', 'python', 'ZODB', 'ZEO', 'tempstorage'):
                    continue

                raise NotImplementedError('%s uses %s without url' % (s, recipe))

            else:
                addbom(url, '')     # XXX detect kind?

        elif recipe == 'slapos.recipe.build:download':
            # slapos.recipe.build:download is often used to download .conf files, but sometimes it is used to download e.g. binaries
            # skip the part, if we can detect that downloaded item is a configuration file
            url = geturl(part)
            if isconf(url):
                continue

            # TODO binary -> parse name/ver
            raise NotImplementedError('%s uses %s with url that does not look like a .conf file: %s' % (s, recipe, url))


        elif recipe.startswith('slapos.recipe.template'):
            url = geturl(part, None)
            if url is not None:
                if isconf(url):
                    continue
                # NOTE binary is not expected with slapos.recipe.template
                raise ValueError('%s uses %s with url that does not look like a .conf file: %s' % (s, recipe, url))
            else:
                # it is an inline= script
                assert 'inline' in part, part


        elif recipe == 'zc.recipe.egg:custom':
            eggpath = part['__buildout_installed__']
            assert len(eggpath.split()) == 1, eggpath  # no spaces inside - just one item
            addbom(eggpath, 'egg')

        elif recipe == 'zc.recipe.egg':
            # XXX sadly zc.recipe.egg neither saves in .installed.cfg information about where the eggs are installed,
            #     nor there are [versions] save.
            #     -> look the egg on the filesystem
            # TODO it is better to fix zc.recipe.egg to save the full information, so that we can build BOM just from .installed.cfg
            eggdir = part['_e']
            eggdev = part['_d']
            eggs = part['eggs'].split()
            for eggname in eggs:
                eggv  = glob('%s/%s-*.egg' % (eggdir, eggname))
                eggv += glob('%s/%s-*.egg' % (eggdev, eggname))
                if len(eggv) == 0:
                    raise ValueError('egg %s not found' % egg)
                if len(eggv) > 1:
                    raise ValueError('egg %s is present multiple times: %s' % (egg, eggv))
                addbom(eggv[0], 'egg')

        else:
            raise NotImplementedError('TODO: add support for recipe %s' % recipe)


    return bom


# geturl returns value of part:url
_missing = object()
def geturl(part, default=_missing):
    url = part.get('url', raw=True)     # raw because e.g. .../ZABBIX%20Latest%20Stable/...
    if url is None:
        if default is not _missing:
            return default
        raise KeyError('section %s has no url' % part)
    return url


def bom_node(XXX):
    1/0
    # TODO bom_node should:
    #
    # 1. retrieve list of software that slapos-node brings with itself
    #    since slapos-node is built via buildout this could be extracted via
    #    node_software from /opt/slapos/.installed.cfg once we fix slapos-node
    #    package to preserve that .installed.cfg in the package itself.
    #
    # 2. retrieve list of base-OS packages that are installed by https://deploy.rapid.space/slapos
    #    This can be computed as the difference in between two sets of base-OS packages:
    #    b) set of base-OS packages after deploy.rapid.space/slapos was run, and
    #    a) set of base-OS packages after minimal base-OS was installed.
    #
    #    this computation could be automated via
    #    - preparing e.g. Debian 11 minimal chroot via debootstrap,
    #    - taking snapshot for list of installed packages via `dpkg -l`,
    #    - then deploying slapos-node in the chroot,
    #    - taking snapshot for the list of installed package again,
    #    - and finally computing the difference.
    #
    # Please see example/ors-bom.txt for "BOM(slapos-node)", "List of software
    # that slapos-node brings with itself" and "List of base-OS packages
    # installed by https://deploy.rapid.space/slapos" sections for what end result could look like.



# namever extracts item name and version from an url/path.
# for example:
#
#     http://www.python.org/ftp/python/2.7.18/Python-2.7.18.tar.xz  ->  ('Python', '2.7.18')
_gitweb_re = re.compile(r'/gitweb/\?p=(?P<name>\w+)\.git;a=snapshot;h=(?P<rev>\w+)')
_github_re = re.compile(r'/github.com/[\w\-]+/(?P<name>[\w\-]+)/archive/(refs/tags/)?(?P<rev>.+)$')
_github_rre= re.compile(r'/raw.githubusercontent.com/[\w\-]+/(?P<name>[\w\-]+)/(?P<rev>[\w\.\-]+)/')
_SF_re     = re.compile(r'/sourceforge.net/.+/(?P<name>[\w\_]+)-(?P<rev>[\w\.]+)/download$')
_go_re     = re.compile(r'/golang.org/dl/(?P<name>go)(?P<rev>[\w\.]+).src$')
_osdn_f    = re.compile(r'/osdn.net/frs/redir.php\?f=(?P<f>.+)$')
def namever(url): # -> (name, ver)
    # http://www.ijg.org/files/jpegsrc.v9d.tar.gz  ->  http://www.ijg.org/files/jpegsrc.v9d
    def del_tgztail(s):
        for tail in ('.tgz', '.tar.gz', '.tbz', '.tar.bz2', '.tar.xz', '.tar.lz', '.zip',
                     '-py2.7.egg', '-py2.7-linux-x86_64.egg'):    # FIXME -> re to cover py2/py3 x86/arm ...
            s = removesuffix(s, tail)
        return s
    url = del_tgztail(url)

    name, ver = _namever(url)
    if ver is not None:
        # swig-3.0.12.tar.gz  ->  swig-3.0.12
        # originally from https://sourceforge.net/projects/swig/files/swig/swig-3.0.12/swig-3.0.12.tar.gz/download
        ver = del_tgztail(ver)
        ver = removeprefix(ver, 'v')
    return name, ver

def _namever(url):
    for r in (_gitweb_re, _github_re, _github_rre, _SF_re, _go_re):
        m = r.search(url)
        if m is not None:
            return m.group('name'), m.group('rev')

    m = _osdn_f.search(url)
    if m is not None:
        url = unquote(m.group('f'))

    filename = basename(url)
    # re.rsplit([-_], filename, 1)
    m = re.search('[-_][^-_]*$', filename)
    if m is not None:
        name = filename[:m.start()]
        ver  = filename[m.start()+1:]
        return name, ver

    m = re.search(r'\.v.+$', filename)  # jpegsrc.v9d
    if m is not None:
        name = filename[:m.start()]
        ver  = filename[m.start()+1:]
        return name, ver

    m = re.match(r'(?P<name>.*[^0-9])(?P<rev>[0-9]+)$', filename) # IPAexfont00201
    if m is not None:
        return m.group('name'), m.group('rev')

    raise RuntimeError('Unsupported url: %r' % (url,))


# isconf returns whether url points to data related to configuration file
# (contrary to binary executable), or a simple/small script.
def isconf(url):
    if url.endswith('.conf.in')                 or \
       url.endswith('.cfg')                     or \
       url.endswith('.cfg.in')                  or \
       url.endswith('.cnf.in')                  or \
       url.endswith('.cfg.jinja2.in')           or \
       url.endswith('.jinja2.cfg')              or \
       url.endswith('.jinja2.library')          or \
       url.endswith('.asn')                     or \
       url.endswith('/ltelogs.jinja2.sh')       or \
       '/promise/' in url                       or \
       url.endswith('/templates/wrapper.in')    or \
       url.endswith('/logrotate_entry.in')      or \
       url.endswith('/matplotlibrc.in')         or \
       url.endswith('/goenv.sh.in')             :
        return True
    return False


# removeprefix/removesuffix provides fallbacks similarly-named str methods.
def removeprefix(s, prefix):
    if s.startswith(prefix):
        s = s[len(prefix):]
    return s
def removesuffix(s, suffix):
    if s.endswith(suffix):
        s = s[:len(s)-len(suffix)]
    return s

# ----------------------------------------

# fmt_bom formats BOM into text.
def fmt_bom(bom): # -> str
    outv = []
    def emit(text):
        outv.append(text+'\n')

    # emit BOM grouped by kind
    kinds = set()
    for info in bom.values():
        kinds.add(info.kind)

    for kind in sorted(kinds):
        if kind != '':
            emit('\n>>> %ss:' % kind)
        for bkey in sorted(bom):
            info = bom[bkey]
            if info.kind == kind:
                # TODO autoalign
                emit('%-28s %-10s %s' % (info.name, info.version, info.url))

    return ''.join(outv)

def main():
    if len(sys.argv) != 3  or  sys.argv[1] not in ('software', 'node'):
        print(__doc__, file=sys.stderr)
        sys.exit(2)

    what, arg = sys.argv[1:]
    if what == 'software':
        bom = bom_software(arg)
    elif what == 'node':
        bom = bom_node(arg)

    # print retrieved BOM
    # TODO also consider emitting machine-readable format, e.g. json if run with --json
    # TODO for json, consider schema from https://cyclonedx.org/specification/overview/
    print(fmt_bom(bom))


if __name__ == '__main__':
    main()
