# Copyright (C) 2023  Nexedi SA and Contributors.
#                     Started by Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.


import nxdbom
import pytest

import os
from os.path import dirname


@pytest.mark.parametrize('url,nameok,verok', [
    ('http://www.python.org/ftp/python/2.7.18/Python-2.7.18.tar.xz',  'Python',  '2.7.18'),
    ('http://www.ijg.org/files/jpegsrc.v9d.tar.gz',                   'jpegsrc', '9d'),
    ('https://github.com/nghttp2/nghttp2/archive/v1.40.0.tar.gz',     'nghttp2', '1.40.0'),
    ('https://golang.org/dl/go1.18.9.src',                            'go',      '1.18.9'),
    ('https://github.com/tesseract-ocr/tesseract/archive/refs/tags/4.1.1.tar.gz',  'tesseract', '4.1.1'),
    ('https://raw.githubusercontent.com/zuphilip/ocropy-models/master/en-default.pyrnn.gz',  'ocropy-models', 'master'),
    ('https://sourceforge.net/projects/swig/files/swig/swig-3.0.12/swig-3.0.12.tar.gz/download',  'swig', '3.0.12'),
    ('https://git.savannah.gnu.org/gitweb/?p=config.git;a=snapshot;h=5e531d39;sf=tgz',  'config', '5e531d39'),
    ('/ROOT/develop-eggs/mysqlclient-1.3.12-py2.7-linux-x86_64.egg',  'mysqlclient', '1.3.12'),
    ('https://osdn.net/frs/redir.php?f=ipafonts%2F57330%2FIPAexfont00201.zip', 'IPAexfont', '00201'),
    ('https://osdn.net/frs/redir.php?f=ipafonts%2F51868%2FIPAfont00303.zip',   'IPAfont',   '00303'),
    ('https://osdn.net/frs/redir.php?f=tsukurimashou%2F56948%2Focr-0.2.zip',   'ocr',       '0.2'),
])
def test_namever(url, nameok, verok):
    assert nxdbom.namever(url) == (nameok, verok)

@pytest.mark.parametrize('url', [
    'xxx.conf.in',
    'xxx.cfg',
    'xxx.cfg.in',
    'xxx.cnf.in',
    'xxx.cfg.jinja2.in',
    'xxx.jinja2.cfg',
    'xxx.jinja2.library',
    'xxx.asn',
    'xxx/ltelogs.jinja2.sh',
    'xxx/templates/wrapper.in',
    'xxx/logrotate_entry.in',
    'xxx/matplotlibrc.in',
    'xxx/goenv.sh.in',
    'xxx/promise/yyy',
])
def test_isconf(url):
    assert nxdbom.isconf(url) == True


# ---- BOM software ----

testv = []  # of (build, bomok)
def case1(build, bomok):
    testv.append((build, bomok))

case1("""\
[ncurses]
recipe = slapos.recipe.cmmi
url = http://ftp.gnu.org/gnu/ncurses/ncurses-6.2.tar.gz
""", """\
ncurses                      6.2        http://ftp.gnu.org/gnu/ncurses/ncurses-6.2.tar.gz
""")

for x in ('gcc', 'python', 'ZODB', 'ZEO', 'tempstorage'):
    case1("""\
[%s]
recipe = slapos.recipe.build
""" % x, '')    # empty


case1("""\
[template-logrotate-base]
recipe = slapos.recipe.template:jinja2
url = /srv/slapgrid/slappart47/srv/project/slapos/stack/logrotate/instance-logrotate-base.cfg.in
""", '')    # config ignored


case1("""\
[python-mysqlclient]
recipe = zc.recipe.egg:custom
__buildout_installed__ = /ROOT/develop-eggs/mysqlclient-1.3.12-py2.7-linux-x86_64.egg
""", """
>>> eggs:
mysqlclient                  1.3.12     https://pypi.org/project/mysqlclient/1.3.12/
""")

case1("""\
[slapos-toolbox-dependencies]
recipe = zc.recipe.egg
_d = /ROOT/develop-eggs
_e = /ROOT/eggs
eggs = lxml
        pycurl
        Mako
-- /ROOT/develop-eggs/lxml-4.9.1-py2.7-linux-x86_64.egg/x --
-- /ROOT/develop-eggs/pycurl-7.43.0-py2.7-linux-x86_64.egg/x --
-- /ROOT/eggs/Mako-1.1.4-py2.7.egg/x --
""", """
>>> eggs:
Mako                         1.1.4      https://pypi.org/project/Mako/1.1.4/
lxml                         4.9.1      https://pypi.org/project/lxml/4.9.1/
pycurl                       7.43.0     https://pypi.org/project/pycurl/7.43.0/
""")

# %20 in URL
case1("""\
[zabbix-agent]
recipe = slapos.recipe.cmmi
url = http://downloads.sourceforge.net/project/zabbix/ZABBIX%20Latest%20Stable/2.0.10/zabbix-2.0.10.tar.gz
""", """\
zabbix                       2.0.10     http://downloads.sourceforge.net/project/zabbix/ZABBIX%20Latest%20Stable/2.0.10/zabbix-2.0.10.tar.gz
""")


@pytest.mark.parametrize('build,bomok', testv)
def test_bom_software(tmpdir, build, bomok):
    tmpdir = str(tmpdir)
    build = '-- /ROOT/.installed.cfg --\n' + build
    build = build.replace('/ROOT', tmpdir)
    build = build.replace('/BASE', tmpdir+'/base')
    ar = txtar_parse(build)
    assert ar.comment == ''
    for f, data in ar.files.items():
        assert f.startswith(tmpdir)
        os.makedirs(dirname(f), exist_ok=True)
        with open(f, 'w') as _:
            _.write(data)

    bom = nxdbom.bom_software(tmpdir)
    assert nxdbom.fmt_bom(bom) == bomok


# ---- txtar ----

# txtar_* provide support for archives in txtar format
# https://pkg.go.dev/golang.org/x/tools/txtar#hdr-Txtar_format
class txtar_Archive:
    # .comment  str
    # .files    {}  name -> text
    pass

def txtar_parse(text): # -> txtar_Archive
    comment = ''
    files = {}
    current_file = None # current file | None for comments
    current_text = ''   # accumulator for text of current file
    def flush(next_file):
        nonlocal comment, current_file, current_text
        if current_file is None:
            comment = current_text
        else:
            files[current_file] = current_text

        current_text = ''
        current_file = next_file

    for l in text.splitlines(keepends=True):
        if not l.endswith('\n'):
            l += '\n'   # missing trailing newline on the final line

        if l.startswith('-- ') and l.rstrip().endswith(' --'):
            _ = l.rstrip()
            _ = _.removeprefix('-- ')
            _ = _.removesuffix(' --')
            next_file = _.strip()
            flush(next_file)
            continue

        current_text += l

    flush(None)
    ar = txtar_Archive()
    ar.comment = comment
    ar.files   = files
    return ar
